/*
 * rtl_time.c
 *
 * PPC-specific clock support
 *
 * Copyright (C) 1999 Cort Dougan <cort@fsmlabs.com>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/timex.h>
#include <linux/mc146818rtc.h>

#include <asm/smp.h>
#include <asm/system.h>
#include <asm/io.h>

#include <rtl_core.h>
#include <rtl_time.h>
#include <rtl.h>

static void _decr_uninit (clockid_t clock);
static int _decr_init (clockid_t clock);
static hrtime_t _decr_gettime (struct rtl_clock *);
static int _decr_settimer (struct rtl_clock *, hrtime_t interval);
static int _decr_settimermode (struct rtl_clock *, int mode);
unsigned int decr_intercept(struct pt_regs *);
unsigned long linux_decrs = 0;
unsigned long linux_decrementer_count = 0, rtl_decrementer_count;

extern unsigned long decrementer_count;
void default_handler( struct pt_regs *regs);

MODULE_AUTHOR("Cort Dougan <cort@fsmlabs.com>");
MODULE_DESCRIPTION("Real-Time Linux PPC Timer Module");

struct rtl_clock decr_clock =
{
	_decr_init, _decr_uninit,
	_decr_gettime,
	NULL, /* sethrtime */
	_decr_settimer,
	_decr_settimermode,
	default_handler, /* handler */
	0, /* mode */
};

hrtime_t _decr_gettime (struct rtl_clock *clock)
{
	unsigned long tbu, tbl, tbu1;
	hrtime_t ret;
	/*
	 * Possible bug pointed out by Niklaus Giger where
	 * the timebase goes from ~0x0 to 0x0.  This
	 * catches that case now.  -- Cort
	 */
	do {
		__asm__("mftbu	%2\n\t"
			"mftb 	%0\n\t"
			"mftbu	%1\n\t"
			: "=r" (tbl), "=r" (tbu), "=r" (tbu1) );
	} while (tbu != tbu1);
	ret = tbl + ((unsigned long long)tbu<<32);
	ret *= ((NSECS_PER_SEC/HZ)/linux_decrementer_count);
	return ret;
}

static int _decr_settimer (struct rtl_clock *clock, hrtime_t interval)
{
	unsigned long t, old_decrementer_count = rtl_decrementer_count;

	if (interval)
	{
		/* ns to decr ticks, rounds down -- Cort */
		t = (long)interval / ((NSECS_PER_SEC/HZ)/linux_decrementer_count);
		if ( t < 500 )
		{
			rtl_printf("t too low %lu in _decr_settimer!\n", t);
			return -1;
		}
		if ( t > linux_decrementer_count )
		{
			rtl_printf("t too high %lu in _decr_settimer!\n", t);
			return -1;
		}
		rtl_decrementer_count = t;
	}
	
	{
		unsigned long dval;
		dval = get_dec();
		set_dec(rtl_decrementer_count);
		/*
		 * Make sure linux doesn't lose these ticks.
		 */
		linux_decrs += old_decrementer_count-dval;
		/*linux_decrs = 0;*/
	}
	return 0;
}

static int _decr_settimermode(struct rtl_clock *clock, int mode)
{
	clock->mode = mode;
	return 0;
}

clockid_t rtl_getbestclock (unsigned int cpu)
{
        return &decr_clock;
}

hrtime_t gethrtime(void)
{
	return _decr_gettime(&decr_clock);
}

hrtime_t gethrtimeres(void)
{
	return (NSECS_PER_SEC/HZ)/linux_decrementer_count;
}

int init_module (void)
{
	/* keep a copy of the decrementer count that Linux wants */
	linux_decrementer_count = decrementer_count;
	rtl_init_standard_clocks();
	return 0;
}

void cleanup_module(void)
{
	rtl_cleanup_standard_clocks();
	decrementer_count = linux_decrementer_count;
}

unsigned int decr_intercept(struct pt_regs *regs)
{
	unsigned long dval;
	int d;
	void (*handler)( struct pt_regs *regs) = decr_clock.handler;
	/* keep track of how many decrs linux have accrued */
	linux_decrs += rtl_decrementer_count;
	
	/*
	 * If it's one-shot clear the handler and reset the 
	 * timer.
	 */
	if ( decr_clock.mode == RTL_CLOCK_MODE_ONESHOT )
	{
		decr_clock.mode = RTL_CLOCK_MODE_PERIODIC;
		rtl_decrementer_count = linux_decrementer_count;
		decr_clock.handler = default_handler;
	}
	
	dval = get_dec();
	while ((d = (int)get_dec()) == dval)
		;
	if ( (int)(d + rtl_decrementer_count) < 500 )
	{
		printk("_decr_intercept(): d %08x decr_count %08lx too low: time overrun likely\n",
		       d, rtl_decrementer_count);
		set_dec(rtl_decrementer_count);
	}
	else
		set_dec(d + rtl_decrementer_count);
	if ( handler != default_handler )
		handler(regs);

	/* make sure the linux timer interrupt handler is called */
	if ( linux_decrs >= linux_decrementer_count )
	{
		linux_decrs -= linux_decrementer_count;
		rtl_local_pend_vec(0,0);
	}
	return 0;
}

static int _decr_init (clockid_t clock)
{
	rtl_decrementer_count = linux_decrementer_count;
	decrementer_count = 0;	/* we're taking over the decr now */
	if ( rtl_request_local_irq(0, decr_intercept, 0) )
	{
		printk("Failed getting local timer irq\n");
		return -1;
	}
	set_dec(rtl_decrementer_count);	  
	return 0;
}

static void _decr_uninit (clockid_t clock)
{
	rtl_free_local_irq(0,0);
	decrementer_count = linux_decrementer_count;
	clock->handler = RTL_CLOCK_DEFAULTS.handler;
}
